<?php
​
/**
 * Copyright © 2016 Filoblu S.r.l. All rights reserved.
 */
​
namespace FiloBlu\Esb\Model\Rest;
​
use Exception;
use FiloBlu\Esb\Api\Data\AcknowledgeInterface;
use FiloBlu\Esb\Api\ExternalShipmentInterface;
use FiloBlu\Esb\Helper\LoggerProvider;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Phrase;
use Magento\InventoryApi\Api\Data\StockInterface;
use Magento\InventorySales\Model\ResourceModel\StockIdResolver;
use Magento\InventorySalesApi\Api\Data\SalesChannelInterface;
use Magento\InventorySourceSelectionApi\Api\Data\InventoryRequestInterfaceFactory;
use Magento\InventorySourceSelectionApi\Api\Data\ItemRequestInterfaceFactory;
use Magento\InventorySourceSelectionApi\Api\GetDefaultSourceSelectionAlgorithmCodeInterface;
use Magento\InventorySourceSelectionApi\Api\SourceSelectionServiceInterface;
use Magento\Sales\Api\Data\OrderExtensionFactory;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Api\Data\ShipmentExtensionInterface;
use Magento\Sales\Api\Data\ShipmentInterface;
use Magento\Sales\Api\ShipmentRepositoryInterface;
use Magento\Sales\Exception\CouldNotShipException;
use Magento\Sales\Model\Order\ShipmentFactory;
use Magento\Sales\Model\OrderRepository;
use Magento\Shipping\Model\ShipmentNotifier;
use Magento\Store\Model\StoreManagerInterface;
use Monolog\Logger;
​
/**
 * REST Api to save external shipments
 */
class ExternalShipment implements ExternalShipmentInterface
{
​
    /**
     * @var OrderRepository
     */
    protected $_orderRepository;
​
    /**
     * @var ShipmentFactory
     */
    protected $_shipmentFactory;
​
    /**
     * @var SearchCriteriaBuilder
     */
    protected $_searchCriteriaBuilder;
​
    /**
     * @var ShipmentFactory
     */
    protected $_shipmentNotifier;
​
    /**
     * @var ScopeConfigInterface
     */
    protected $_scopeConfig;
​
    /**
     * @var OrderExtensionFactory
     */
    protected $_orderExtensionFactory;
​
    /**
     * @var Logger
     */
    protected $_logger;
​
    /**
     * @var AcknowledgeInterface
     */
    protected $_response;
​
    /**
     * @var ShipmentRepositoryInterface
     */
    private $shipmentRepository;
    /**
     * @var ProductMetadataInterface
     */
    private $productMetadata;
​
    /**
     * Init plugin
     *
     * @param ScopeConfigInterface $scopeConfig
     * @param OrderRepository $orderRepository
     * @param ShipmentFactory $shipmentFactory
     * @param OrderExtensionFactory $orderExtensionFactory
     * @param ShipmentNotifier $shipmentNotifier
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param LoggerProvider $loggerProvider
     * @param ShipmentRepositoryInterface $shipmentRepository
     * @param ProductMetadataInterface $productMetadata
     * @param AcknowledgeInterface $response
     */
    public function __construct(
        ScopeConfigInterface $scopeConfig,
        OrderRepository $orderRepository,
        ShipmentFactory $shipmentFactory,
        OrderExtensionFactory $orderExtensionFactory,
        ShipmentNotifier $shipmentNotifier,
        SearchCriteriaBuilder $searchCriteriaBuilder,
        LoggerProvider $loggerProvider,
        ShipmentRepositoryInterface $shipmentRepository,
        ProductMetadataInterface $productMetadata,
        AcknowledgeInterface $response
    ) {
        $this->_scopeConfig = $scopeConfig;
        $this->_orderRepository = $orderRepository;
        $this->_shipmentFactory = $shipmentFactory;
        $this->_orderExtensionFactory = $orderExtensionFactory;
        $this->_shipmentNotifier = $shipmentNotifier;
        $this->_searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->_logger = $loggerProvider->getLogger();
        $this->_response = $response;
        $this->shipmentRepository = $shipmentRepository;
        $this->productMetadata = $productMetadata;
    }
​
    /**
     * Save External Shipment data. Save PDF file and additional data on Sales Order.
     *
     * @param ShipmentInterface $shipment_data
     * @return AcknowledgeInterface
     * @throws Exception
     */
    public function save(ShipmentInterface $shipment_data)
    {
        $shipmentExtensionAttributes = $shipment_data->getExtensionAttributes();
​
        if (null === $shipmentExtensionAttributes || !$shipmentExtensionAttributes->getOrderIncrementId()) {
            $message = __('Cannot find order to ship');
            $this->_logger->critical($message->render());
            throw new CouldNotShipException(new Phrase($message->render()));
        }
​
        $orderIncrementId = $shipmentExtensionAttributes->getOrderIncrementId();
        $carrierCode = $shipmentExtensionAttributes->getCarrierCode();
        if (!$carrierCode) {
            $carrierCode = 'ups';
        }
        $carrierTitle = $shipmentExtensionAttributes->getCarrierTitle();
        if (!$carrierTitle) {
            $carrierTitle = 'UPS';
        }
        $searchCriteria = $this->_searchCriteriaBuilder
            ->addFilter('increment_id', $orderIncrementId, 'eq')->create();
        $orderList = $this->_orderRepository->getList($searchCriteria);
        if (count($orderList) !== 1) {
            $message = __('Cannot find order %1', $orderIncrementId);
            $this->_logger->critical($message->render());
            throw new CouldNotShipException(new Phrase($message->render()));
        }
        try {
            $order = $orderList->getFirstItem();
            //In magento 2.2 empty item array is not handled any more. Build array with qty_to_ship.
            $items = [];
            foreach ($order->getAllItems() as $orderItem) {
                if (!$orderItem->getQtyToShip() || $orderItem->getIsVirtual()) {
                    continue;
                }
                $qtyShipped = $orderItem->getQtyToShip();
                $items[$orderItem->getId()] = $qtyShipped;
            }
            $tracks = $shipment_data->getTracks();
            $shipmentParameters = null;
            $number = '';
            if ($tracks) {
                $track = current($tracks);
                try {
                    $number = $track->getTrackNumber();
                    $shipmentParameters = [
                        [
                            'order_id' => $order->getId(),
                            'carrier_code' => $carrierCode,
                            'title' => $carrierTitle,
                            'number' => $number
                        ]
                    ];
                } catch (Exception $e) {
                    $this->_logger->error($e->getMessage(), ['exception', $e]);
                }
            }
​
            $shipment = $this->_shipmentFactory->create(
                $order,
                $items,
                $shipmentParameters
            );
​
            $shipment->register();
            $shipment->getOrder()->setIsInProcess(true);
            $this->addMultiStockInformationIfNeeded($shipment, $order);
            $this->shipmentRepository->save($shipment);
            $order->addStatusHistoryComment(__('Order has been shipped: %1 ', $number), 'shipped');
            $order->save();
            $shipment = $this->shipmentRepository->get($shipment->getEntityId());
            $this->_shipmentNotifier->notify($shipment);
​
            $this->_response->setStatus('success');
            $this->_response->setMessage(__('Order %1 has been shipped:  %2', $order->getIncrementId(), $number));
            $this->_response->setFlow('update_shipment');
            $this->_response->setOrderIncrementId($order->getIncrementId());
            return $this->_response;
        } catch (Exception $e) {
            $this->_logger->critical($e->getMessage(), ['exception', $e]);
            throw $e;
        }
    }
​
    /**
     * @param ShipmentInterface $shipment
     * @param OrderInterface $order
     * @throws LocalizedException
     * @throws NoSuchEntityException
     * @throws Exception
     */
    protected function addMultiStockInformationIfNeeded(ShipmentInterface $shipment, OrderInterface $order)
    {
        if (version_compare($this->productMetadata->getVersion(), '2.3.0', '<')) {
            return;
        }
​
        $order->getStoreId();
        $objectManager = ObjectManager::getInstance();
​
        $sourceSelection = $objectManager->create(GetDefaultSourceSelectionAlgorithmCodeInterface::class);
​
        $itemRequestFactory = $objectManager->create(ItemRequestInterfaceFactory::class);
        $inventoryRequestFactory = $objectManager->create(InventoryRequestInterfaceFactory::class);
        $sourceSelectionService = $objectManager->create(SourceSelectionServiceInterface::class);
        $storeManager = $objectManager->create(StoreManagerInterface::class);
        $websiteId = $storeManager->getStore($order->getStoreId())->getWebsiteId();
​
        /** @var StockIdResolver $stockIdResolver */
        $stockIdResolver = $objectManager->create(StockIdResolver::class);
        /** @var StockInterface $stock */
        $stockId = $stockIdResolver->resolve(SalesChannelInterface::TYPE_WEBSITE, $storeManager->getWebsite($websiteId)->getCode());
        $algorithmCode = $sourceSelection->execute();

        foreach ($order->getItems() as $item) {
            $requestItem = $itemRequestFactory->create([
                'sku' => $item->getSku(),
                'qty' => $item->getQtyInvoiced()
            ]);
            $inventoryRequest = $inventoryRequestFactory->create([
                'stockId' => $stockId,
                'items' => [$requestItem]
            ]);
            $sourceSelectionResult = $sourceSelectionService->execute(
                $inventoryRequest,
                $algorithmCode
            );
​
            if (!$sourceSelectionResult->isShippable()) {
                throw new \RuntimeException('No quantity available for shipment');
            }
​
            foreach ($sourceSelectionResult->getSourceSelectionItems() as $product) {
                if ($product->getQtyAvailable() >= $item->getQtyInvoiced()) {
                    /** @var ShipmentExtensionInterface $extensionAttributes */
                    $extensionAttributes = $shipment->getExtensionAttributes();
                    $extensionAttributes->setSourceCode($product->getSourceCode());
                    $shipment->setExtensionAttributes($extensionAttributes);
                    return;
                }
            }
        }
​
        throw new \RuntimeException('No quantity available for shipment');
    }
}